"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.visitFnResourceBody = void 0;
const typescript_1 = __importDefault(require("typescript"));
const base_1 = require("@plutolang/base");
const imports_1 = require("./imports");
const dep_resolve_1 = require("./dep-resolve");
const constants_1 = require("./constants");
const utils_1 = require("./utils");
/**
 * Construct the FnResource and detect the access relationship within the body of FnResource.
 * @param fnNode The function expression
 * @param fnResName The function name, same with the resource name.
 */
function visitFnResourceBody(fnNode, checker, fnResName) {
    const fnScope = [fnNode.getStart(), fnNode.getEnd()];
    // Visit this function body, and detect the variable accessing, function calling and resource using in the access chain.
    const detectResult = detectAccessChain(fnNode, fnScope, checker, fnResName, 0);
    if (!detectResult) {
        throw new Error("This detection should not be undefined.");
    }
    const fnLoc = (0, utils_1.getLocationOfNode)(fnNode, 0);
    if (new Set(detectResult?.depLocs.concat(fnLoc).map((loc) => loc.file)).size != 1) {
        throw new Error(`Currently, Pluto only supports a single file.`);
    }
    const depLocs = removeDuplicateLocs(detectResult.depLocs.concat(fnLoc));
    // Get the import dependencies for the function.
    const bodyImportElements = (0, dep_resolve_1.resolveImportDeps)(fnNode.getSourceFile(), (0, imports_1.buildImportStore)(fnNode.getSourceFile()), fnNode);
    let importElements = bodyImportElements.concat(detectResult.importElements);
    importElements = importElements.filter((obj1, obj2, arr) => arr.findIndex((val) => val.name === obj1.name) === obj2);
    // Consturct the Fn Resource.
    const resourceVarInfo = {
        varName: fnResName,
        resourceConstructInfo: {
            constructExpression: constants_1.FN_RESOURCE_TYPE_NAME,
            importElements: importElements,
            locations: depLocs,
        },
    };
    const relatInfos = removeDuplicateRelatInfos(detectResult.resRelatInfos);
    return {
        resourceVarInfos: [resourceVarInfo],
        resourceRelatInfos: relatInfos,
    };
}
exports.visitFnResourceBody = visitFnResourceBody;
function combineDetectFnAccessChainResults(...results) {
    const resRelatInfos = [];
    const importElements = [];
    const depLocs = [];
    results.forEach((result) => {
        if (!result)
            return;
        resRelatInfos.push(...result.resRelatInfos);
        importElements.push(...result.importElements);
        depLocs.push(...result.depLocs);
    });
    return {
        resRelatInfos,
        importElements,
        depLocs,
    };
}
/**
 * Based on the type of current node, detect three types of accesses:
 *   1. If this node is an identifier, check if it is a constant variable located outside of this node scope.
 *      If yes, add its declation location into the dependency list.
 *   2. If this node is a call expression, check if it is a function call located outside of this node scope.
 *      If yes, add its declaration location into the dependency list.
 *   3. If this node is a variable statement or expression statement, check if it is a resource variable.
 *      If yes, establish a relationship between the root FnResource in the chain and the target resource.
 * @param curNode The node need to be checked.
 * @param curScope The scope in which the node locates.
 */
function detectAccessChain(curNode, curScope, checker, rootFnName, depth) {
    const curFile = curNode.getSourceFile().fileName;
    if (curFile.indexOf("node_modules") !== -1)
        return;
    let combinedRes;
    if (typescript_1.default.isIdentifier(curNode)) {
        // Might be accessing a constant variable located outside of scope.
        combinedRes = detectFnAccessConst(curNode, curScope, checker, rootFnName, depth);
    }
    else if (typescript_1.default.isCallExpression(curNode)) {
        // Might be calling function located outside of scope.
        combinedRes = detectFnAccessFn(curNode, curScope, checker, rootFnName, depth);
    }
    else if (typescript_1.default.isExpressionStatement(curNode) || typescript_1.default.isVariableStatement(curNode)) {
        // Might be calling a resource method.
        combinedRes = detectFnAccessResource(curNode, checker, rootFnName);
    }
    curNode.forEachChild((childNode) => {
        const result = detectAccessChain(childNode, curScope, checker, rootFnName, depth);
        combinedRes = combineDetectFnAccessChainResults(combinedRes, result);
    });
    return combinedRes;
}
/**
 * Check if this node is using resource.
 * This information will be utilized to generate the permission configuration for this FnResource.
 */
function detectFnAccessResource(curNode, checker, fnResName) {
    let propAccessExp;
    // Write operation, e.g. state.set(), queue.push()
    if (typescript_1.default.isExpressionStatement(curNode) &&
        typescript_1.default.isAwaitExpression(curNode.expression) &&
        typescript_1.default.isCallExpression(curNode.expression.expression) &&
        typescript_1.default.isPropertyAccessExpression(curNode.expression.expression.expression)) {
        propAccessExp = curNode.expression.expression.expression;
    }
    else if (typescript_1.default.isVariableStatement(curNode)) {
        // Read operation, e.g. state.get()
        const initExp = curNode.declarationList.declarations[0].initializer;
        if (initExp &&
            typescript_1.default.isAwaitExpression(initExp) &&
            typescript_1.default.isCallExpression(initExp.expression) &&
            typescript_1.default.isPropertyAccessExpression(initExp.expression.expression)) {
            propAccessExp = initExp.expression.expression;
        }
    }
    if (!propAccessExp || !(0, utils_1.isResourceVar)(propAccessExp.expression, checker)) {
        return;
    }
    const accessorName = propAccessExp.expression.getText();
    const symbol = checker.getSymbolAtLocation(propAccessExp);
    if (symbol == undefined) {
        throw new Error("The symbol of this function call is undefined: " + propAccessExp.getText());
    }
    const fnName = checker.symbolToString(symbol);
    const resRelatInfo = {
        fromVarName: fnResName,
        toVarNames: [accessorName],
        type: base_1.arch.RelatType.ACCESS,
        operation: fnName,
        parameters: [],
    };
    return {
        resRelatInfos: [resRelatInfo],
        importElements: [],
        depLocs: [],
    };
}
/**
 * Check if this node is a variable located outside of this node scope.
 * If yes, add its declaration location into the dependency list.
 */
function detectFnAccessConst(curNode, curScope, checker, rootFnName, depth) {
    const symbol = checker.getSymbolAtLocation(curNode);
    // Check if this identifier is a constant variable.
    if (!symbol || !(0, utils_1.isConstVar)(symbol) || (0, utils_1.isResourceVar)(curNode, checker)) {
        return;
    }
    const type = checker.getTypeOfSymbol(symbol);
    if (!type.isLiteral()) {
        throw new Error("Currently, Pluto only supports accessing constant variables with literal values that are outside the scope of a function.");
    }
    // If this is a constant variable that is defined inside the scope of this function, we can ignore it.
    const declStat = getSymbolDeclStatement(symbol);
    if (declStat.getSourceFile().fileName.indexOf("node_modules") !== -1 ||
        (declStat.getSourceFile().fileName === curNode.getSourceFile().fileName &&
            declStat.getStart() >= curScope[0] &&
            declStat.getEnd() <= curScope[1])) {
        return;
    }
    // Since the constant variable is declared outside of the current scope,
    // we need to identify the import dependencies required for the declaration of the constant variable.
    const importElements = (0, dep_resolve_1.resolveImportDeps)(declStat.getSourceFile(), (0, imports_1.buildImportStore)(declStat.getSourceFile()), declStat);
    let combinedRes = { resRelatInfos: [], importElements, depLocs: [] };
    const declLoc = (0, utils_1.getLocationOfNode)(declStat, depth + 1);
    const declScope = [declStat.getStart(), declStat.getEnd()];
    combinedRes.depLocs.push(declLoc);
    const childResult = detectAccessChain(declStat, declScope, checker, rootFnName, depth + 1);
    combinedRes = combineDetectFnAccessChainResults(combinedRes, childResult);
    return combinedRes;
}
/**
 * Check if this node is an expression that calls a function located outside the scope of this node.
 * If yes, add its declaration location into the dependency list.
 */
function detectFnAccessFn(curNode, curScope, checker, rootFnName, depth) {
    if (!typescript_1.default.isCallExpression(curNode)) {
        return;
    }
    const fnExpression = curNode.expression;
    const symbol = checker.getSymbolAtLocation(fnExpression);
    if (!symbol) {
        return;
    }
    const declStat = getSymbolDeclStatement(symbol);
    if (declStat.getSourceFile().fileName.indexOf("node_modules") !== -1 ||
        (declStat.getSourceFile().fileName == curNode.getSourceFile().fileName &&
            declStat.getStart() > curScope[0] &&
            declStat.getEnd() < curScope[1])) {
        // If the function being called is a library function or if it is defined within the scope of this function, we can disregard it.
        return;
    }
    if (typescript_1.default.isPropertyAccessExpression(fnExpression)) {
        let headNode = fnExpression.expression;
        while (typescript_1.default.isPropertyAccessExpression(headNode)) {
            headNode = headNode.expression;
        }
        if ((0, utils_1.isResourceVar)(headNode, checker)) {
            return;
        }
        throw new Error("Currently, Pluto doesn't support property calling.");
    }
    // Since the function is declared outside of the current scope,
    // we need to identify the import dependencies required for the declaration of the function.
    const importElements = (0, dep_resolve_1.resolveImportDeps)(declStat.getSourceFile(), (0, imports_1.buildImportStore)(declStat.getSourceFile()), declStat);
    let combinedRes = {
        resRelatInfos: [],
        importElements: importElements,
        depLocs: [],
    };
    const declLoc = (0, utils_1.getLocationOfNode)(declStat, depth + 1);
    const declScope = [declStat.getStart(), declStat.getEnd()];
    combinedRes.depLocs.push(declLoc);
    const childResult = detectAccessChain(declStat, declScope, checker, rootFnName, depth + 1);
    combinedRes = combineDetectFnAccessChainResults(combinedRes, childResult);
    return combinedRes;
}
function getSymbolDeclStatement(symbol) {
    const symbolDeclaration = symbol.valueDeclaration || symbol.declarations?.[0];
    if (symbolDeclaration == undefined) {
        throw new Error("Cannot found the declaration of symbol: " + symbol.name);
    }
    let parNode = symbolDeclaration;
    while (!typescript_1.default.isStatement(parNode)) {
        parNode = parNode.parent;
    }
    return parNode;
}
function removeDuplicateLocs(oldLocs) {
    const newLocs = [];
    for (const oldLoc of oldLocs) {
        let existed = false;
        for (const newLoc of newLocs) {
            if (oldLoc.file != newLoc.file || oldLoc.start != newLoc.start || oldLoc.end != newLoc.end) {
                continue;
            }
            existed = true;
            newLoc.depth = Math.min(newLoc.depth, oldLoc.depth);
        }
        if (!existed)
            newLocs.push(oldLoc);
    }
    return newLocs;
}
/**
 * Remove duplicate relationship information.
 */
function removeDuplicateRelatInfos(oldRelatInfos) {
    const existedKeys = new Set();
    const newRelatInfos = [];
    for (const relatInfo of oldRelatInfos) {
        const key = `${relatInfo.operation}-${relatInfo.fromVarName}-${relatInfo.toVarNames.join("")}`;
        if (existedKeys.has(key))
            continue;
        newRelatInfos.push(relatInfo);
        existedKeys.add(key);
    }
    return newRelatInfos;
}
//# sourceMappingURL=visit-fn-res.js.map